/*
Copyright (c) 2011, Julian Lyndon-Smith (julian+maia@dotr.com)
http://www.dotr.com
All rights reserved.

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction except as noted below, including without limitation
the rights to use,copy, modify, merge, publish, distribute,
and/or sublicense, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

 The Software and/or source code cannot be copied in whole and
 sold without meaningful modification for a profit.

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with
 the distribution.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

using Progress.Lang.*.
using dotr.Maia.Model.*.

/** maia core 
 * this class is responsible for all of the building of the files of a template. for each supplied object, 
 * read the template in, and then parse it for directives. Each directive in the template is a method in the builders
 * defined in the template. 
 * @author Julian Lyndon-Smith (julian+maia@dotr.com) 
 */

routine-level on error undo, throw.

class dotr.Maia.Builder.ObjectBuilder: 
  
  def public static property Instance as dotr.Maia.Builder.ObjectBuilder no-undo
    get():
      if not valid-object(Instance) then Instance = new dotr.Maia.Builder.ObjectBuilder().
      return instance.
    end get . private set.
      
  /** temp-table used to store options for the build */  
  def temp-table TTBuilderOption no-undo
    field OptionName  as char
    field OptionValue as char
    
    index OptionName is primary unique
      OptionName.

  constructor private ObjectBuilder():
  end constructor.
  
  /** get the object of the supplied parameter
   * @param p_ObjectConfig : name of the object to buidl
   */

  method public void BuildObject(p_ObjectConfig as char ):
    BuildObject(ObjectConfig:Get(p_ObjectConfig)).
    return. 
  end method.
  
  /** build supplied object
   * this is the main component of the ObjectBuilder. All templates are read and parsed here
   * and the appropriate directive invoked
   * @param p_ObjectConfig : object to buidl
   */
  method public void BuildObject(p_ObjectConfig as ObjectConfig ):
    def var lv_Custom as longchar no-undo.
    def var lv_Base   as longchar no-undo.
    
    def var lv_Builder     as char no-undo.
    def var lv_File        as char no-undo.
    def var lv_FileMD5Hash as char no-undo.
    def var lv_OrigMD5Hash as char no-undo.
    def var lv_Class       as char no-undo.
    def var lv_Info        as char no-undo. 
    
    def var lv_Startindex as int no-undo.
    def var lv_Endindex   as int no-undo.
    def var lv_StartLine  as int no-undo.
    def var lv_i          as int no-undo.
    
    def var Directive1 as Directive no-undo.
    def var Builder1   as dotr.Maia.Builder.Basic   no-undo.
    
    assign p_ObjectConfig:ObjectBuilder    = this-object.
    
    /** choose the dictdb  */
    if p_ObjectConfig:logicalDB ne ldbname("DICTDB") and p_ObjectConfig:logicalDB ne "" then
    do:
      delete alias DICTDB.
      create alias DICTDB for database value(if p_ObjectConfig:logicalDB eq "" then GetOptionValue("DB")
                                                                               else p_ObjectConfig:logicalDB).
    end.
    
    (new dotr.Maia.Builder.DBProperties()):GetDBProperties(p_ObjectConfig).
    
    /** run through all templates defined for this object */
    templateloop:
    do lv_i = 1 to p_ObjectConfig:NumTemplates on error undo, throw:

      assign lv_Custom = ""
             lv_Base   = ""
             lv_Base    = GetTemplate(substitute("&1/&2",dotr.Maia.CurrentProject:TemplateDirectory,p_ObjectConfig:Template[lv_i]:TemplateName))
             lv_Base    = GetOptions(lv_Base).

      if not p_ObjectConfig:Template[lv_i]:Build then next templateloop.
     
      /** decide which builder to use for this template */
      assign lv_Builder = GetOptionValue("Builder").
      
      /** instantiate the builder */
      Builder1 = dynamic-new string(lv_Builder) (p_ObjectConfig).
         
      /** oops */
      if not Builder1:BuilderAvailable then 
      do:
        message "Unable to load builder. This may be caused by having no UI properties defined for a UI builder" lv_Builder view-as alert-box warning.
        next templateloop.
      end.
            
      /** set options for this build. if the template does not supply an option, then use a default value */
      if GetOptionValue("NoSuffix") eq "" then SetDefaultOption("ClassSuffix",entry(1,p_ObjectConfig:Template[lv_i]:TemplateName,".")).
      
      assign lv_Class = entry(num-entries(p_ObjectConfig:ClassName,"!"),p_ObjectConfig:ClassName,"!").
      
      SetDefaultOption("AppPackage", dotr.Maia.CurrentProject:Name).

      SetDefaultOption("BaseName",lv_class).
      SetDefaultOption("Builder","dotr.Maia.Builder.Basic").
      SetDefaultOption("BuildPackage", Builder1:BuildPackage).
      SetDefaultOption("BuildDirectory", Builder1:BuildDirectory).

      SetDefaultOption("DB", p_ObjectConfig:LogicalDb).

      SetDefaultOption("KeyFieldDatatype",p_ObjectConfig:KeyFieldDatatype).
      SetDefaultOption("KeyFieldName",p_ObjectConfig:KeyFieldName).
      SetDefaultOption("KeyProperty",p_ObjectConfig:KeyProperty).

      SetDefaultOption("ClassName",substitute("&1&2",lv_class,GetOptionValue("ClassSuffix"))).

      SetDefaultOption("StompServer","localhost").
      
      SetDefaultOption("TableName",p_ObjectConfig:TableName).


      /** the xml file is the config for the object */
      SetOption("BuildFileName",dotr.Maia.CurrentProject:GetConfigFile(p_ObjectConfig:ClassName)).
      SetOption("BuildDate",string(now)).
            
      assign lv_File    = substitute("&1/&2.cls",Builder1:BuildDirectory,GetOptionValue("ClassName")).

      assign p_ObjectConfig:OutputFile                = lv_File
             p_ObjectConfig:CurrentClassName          = GetOptionValue("ClassName")
             p_ObjectConfig:Template[lv_i]:OutputFile = lv_File.
  
      /** if the output file already exists, check the MD5 Hash value to see if there have been any modifications since it was last built */
      if search(p_ObjectConfig:OutputFile) ne ? then 
      do:
        /** check to see if we should overwrite this file, as it already exists */
        if not p_ObjectConfig:Template[lv_i]:BuildIfExists then next templateloop.
        
        assign file-info:file-name = search(p_ObjectConfig:OutputFile).
    
        copy-lob from file file-info:full-pathname to lv_Custom.
    
        /** try to get original MD5 hash */
        assign lv_info        = '@BuildInfo(MD5Hash="'
               lv_StartIndex  = index(lv_Custom,lv_info)
               lv_OrigMD5Hash = if lv_StartIndex gt 0 then substr(lv_Custom,lv_StartIndex + length(lv_info),24)
                                                      else p_ObjectConfig:Template[lv_i]:MD5Hash .
        
        assign lv_StartIndex = index(lv_Custom,"@BuildInfo(Source=").
        
        /** remove the buildinfo from md5 calculation */
        if lv_StartIndex gt 0 then 
        do:
          assign substr(lv_Custom,lv_StartIndex,length(lv_custom) - lv_StartIndex + 1) = ""
                 lv_Custom = right-trim(lv_Custom,"~n").
        end.
        
        assign lv_FileMD5Hash = base64-encode(md5-digest(lv_Custom,"Maia")).
        
       if lv_OrigMD5Hash ne lv_FileMD5Hash then
        do:
          message substitute("Warning: &1 has an invalid MD5 value.~nThis indicates that the file has been modified",p_ObjectConfig:OutputFile) skip
                  "If you continue, any changes made to this file will be overwritten and lost" skip
                  "Do you want to continue building this file ?"
                  view-as alert-box question buttons yes-no update lv_Continue as logical.
          if not lv_Continue then next templateloop.
          
          /** make a backup of the modified file */
          copy-lob from lv_Custom to file substitute("&1.&2",file-info:full-pathname,interval(now,1/1/1970,"milliseconds")).

        end.
        
      end.
      
      /** pass the template into the startbuild method for modifications to the template */
      assign lv_base = Builder1:StartBuild(lv_Base).
      
      /** loop through directives in the template 
        * a directive is started by a "~{&!" and ended with a "!}" */  
      do while true on error undo, throw:
        assign lv_StartIndex = index(lv_Base,"~{&!").
        
        /** no more directives to parse */
        if lv_StartIndex le 0 then leave. 
       
        assign lv_StartLine = r-index(lv_Base,"~n",lv_StartIndex) + 1
               lv_EndIndex  = index(lv_Base,"!}",lv_StartIndex) + 2.
       
        /** create a new dirctive object */
        Directive1 = new Directive(string(substr(lv_Base,lv_StartIndex,lv_EndIndex - lv_StartIndex))).
       
        /** invoke the directive method in the builder */
        dynamic-invoke(Builder1,Directive1:DirectiveName,Directive1) no-error.
        
        assign Directive1:Data = replace(Directive1:Data,"~r","").
        
        /** trim the code so that we don't insert blank lines, and the supplied code lines up with 
          * the previous statements in the template */
        if Directive1:Data eq "" and Directive1:TrimCode then
        do:
          assign lv_StartLine = MAX(lv_StartLine - 1,1)
                 lv_EndIndex  = index(lv_Base,"~n",lv_endIndex).
          assign substr(lv_Base,lv_StartLine,(lv_EndIndex - lv_StartLine) ) = "".
      
        end.
        
        else
        do:
          assign substr(lv_Base,lv_StartIndex,(lv_EndIndex - lv_StartIndex)) = ""
                 substr(lv_Base,lv_StartIndex,0)                           = replace(Directive1:Data,"~n","~n" + fill(" ",lv_StartIndex - lv_StartLine)).
        end.  
        
        finally:
          delete object Directive1 no-error.   
        end finally.                   
      end.
      
      /** loop through all builder options, replacing the text in the template <<OptionName>> with the option value */ 
      for each TTBuilderOption no-lock:
        assign lv_Base = replace(lv_Base,substitute("<<&1>>",TTBuilderOption.OptionName),TTBuilderOption.OptionValue).
      end.

      /** pass modified template data to endbuild method */
      assign lv_base = Builder1:EndBuild(lv_Base).
            
      assign lv_base        = right-trim(lv_base,"~n")
             lv_FileMD5Hash = base64-encode(md5-digest(lv_base,"Maia")).

      /** if the template md5 hash is the same as the new md5 hash, then the file hasn't changed after a build
       *  check to see if the configuration still requires a new build . */ 
      
      if search(p_ObjectConfig:OutputFile) ne ? and p_ObjectConfig:Template[lv_i]:MD5Hash eq lv_FileMD5Hash and not p_ObjectConfig:Template[lv_i]:BuildIfSame then next. 

      p_ObjectConfig:Template[lv_i]:MD5Hash = lv_FileMD5Hash.
      
      if lv_base eq ? or lv_base eq "" then message "oh dear" view-as alert-box. /** something has gone wrong with building the template */
      
      else
      do:
        if search(p_ObjectConfig:OutputFile) eq ? then /** we need to create a dummy file so file-info:file-name does not crap out */
        do:
          output to value(p_ObjectConfig:OutputFile).
          output close.
        end.
        
        assign file-info:file-name = p_ObjectConfig:OutputFile
               lv_base             = lv_base + "~n" + substitute('@BuildInfo(Source="&1").~n@BuildInfo(Date="&2").~n@BuildInfo(MD5Hash="&3").',
                                                                  GetOptionValue("BuildFileName"),
                                                                  GetOptionValue("BuildDate"),
                                                                  lv_FileMD5Hash).
        copy-lob lv_Base to file file-info:full-pathname.
      end.
      
    end.

    return.
    
    finally:
    
    end finally.
                    
  end method.
  
  /** reads the template in
   *  after reading, check for "include" directives. If found, insert their contents into the data structure as well.
   * @param p_BaseTemplate : template to read
   * @return longchar : template contents
   */
  
  method private longchar GetTemplate(p_BaseTemplate as char):
    def var lv_Data     as longchar no-undo.
    def var lv_Template as longchar no-undo.
  
    def var Directive1 as Directive no-undo.
    
    def var lv_StartLine  as int no-undo.
    def var lv_Startindex as int no-undo.
    def var lv_Endindex   as int no-undo.
    def var lv_i          as int no-undo.
    
    def var lv_fulltemplate as char no-undo.
    
    assign lv_fulltemplate = p_BaseTemplate.
    
    if search(lv_fullTemplate) eq ? then undo, throw new AppError(substitute("Cannot find template [&1]",lv_fullTemplate),0).
    
    assign file-info:file-name = lv_fullTemplate.
         
    copy-lob from file file-info:full-pathname to lv_Data.
    
    /** loop looking for include directives */
    do while true on error undo, throw:
      assign lv_StartIndex = index(lv_Data,"~{&! Includes").
      if lv_StartIndex le 0 then leave.

      assign lv_StartLine = r-index(lv_Data,"~n",lv_StartIndex) + 1
             lv_EndIndex  = index(lv_Data,"!}",lv_StartIndex) + 2.
     
      /** create new directive */
      Directive1 = new Directive(string(substr(lv_Data,lv_StartIndex,lv_EndIndex - lv_StartIndex))).
 
      assign substr(lv_Data,lv_StartIndex,(index(lv_Data,"~n",lv_StartIndex) + 1) - lv_StartIndex) = "".
       
      /** insert included template contents into data */
      do lv_i = 1 to extent(Directive1:ParameterData):
        assign Directive1:Data = Directive1:Data + "~n" + GetTemplate(Directive1:ParameterName[lv_i]).
      end.
      
      assign substr(lv_Data,lv_StartIndex,0) = replace(Directive1:Data,"~n","~n" + fill(" ",lv_StartIndex - lv_StartLine)).  
                    
      finally:
        delete object Directive1 no-error.   
      end finally.                   
    end.
    
    return replace(lv_Data,"~r","").
  
  end method.

  /** returns the Builder option value
   * @param p_Property : name of option
   * @return char : value of option
   */
  
  method public char GetOptionValue(p_Property as char):
    find TTBuilderOption where TTBuilderOption.OptionName eq p_Property no-lock no-error.
    
    return if avail TTBuilderOption then TTBuilderOption.OptionValue else "".  
  end method.

  /** sets a default builder option value
   * if option already exists, do not overwrite it
   * @param p_Property : name of option to set
   * @param p_Value : default value to set
   */

  method public void SetDefaultOption(p_Property as char,p_Value as char):
    find TTBuilderOption where TTBuilderOption.OptionName eq p_Property no-lock no-error.
    
    if avail TTBuilderOption then return.
    SetOption(p_Property,p_Value).
  end method.
  
  /** sets a builder option value
   * @param p_Property : name of option to set
   * @param p_Value : default value to set
   */

  method public void SetOption(p_Property as char,p_Value as char):
    if p_Property eq "" then return.
    
    find TTBuilderOption where TTBuilderOption.OptionName eq p_Property no-lock no-error.
    
    if not avail TTBuilderOption then create TTBuilderOption.
    
    assign TTBuilderOption.OptionName  = p_Property
           TTBuilderOption.OptionValue = p_Value.
  end method.
  
  /** reads all Option directives in a template
   * @param p_Data : template contents
   * @return longchar : modified template
   */
  
  method private longchar GetOptions(p_Data as longchar):
    def var Directive1 as Directive no-undo.
    
    def var lv_Startindex as int no-undo.
    def var lv_Endindex   as int no-undo.
    def var lv_i          as int no-undo.

    empty temp-table TTBuilderOption.
        
    /** loop through all defineoption directives */
    do while true on error undo, throw:
      assign lv_StartIndex = index(p_Data,"~{&! DefineOption").
      if lv_StartIndex le 0 then leave.
     
      assign lv_EndIndex  = index(p_Data,"!}",lv_StartIndex) + 2.

      Directive1 = new Directive(string(substr(p_Data,lv_StartIndex,lv_EndIndex - lv_StartIndex))).
     
      assign substr(p_Data,lv_StartIndex,(index(p_Data,"~n",lv_StartIndex) + 1) - lv_StartIndex) = "".
      
      do lv_i = 1 to extent(Directive1:ParameterData):
        SetDefaultOption(Directive1:ParameterName[lv_i],Directive1:ParameterData[lv_i]).
      end.
      
      finally:
        delete object Directive1 no-error.   
      end finally.                   
    end.
    
    return p_Data.
  
  end method.

end class.