/*
Copyright (c) 2011, Julian Lyndon-Smith (julian+maia@dotr.com)
http://www.dotr.com
All rights reserved.

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction except as noted below, including without limitation
the rights to use,copy, modify, merge, publish, distribute,
and/or sublicense, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

 The Software and/or source code cannot be copied in whole and
 sold without meaningful modification for a profit.

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with
 the distribution.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

using Progress.Lang.*.
using dotr.Maia.Model.*.

/** crud library for Project 
 * this class is responsible for all crud activity on the Project model. It serialises and deserialises the 
 * config model and all related children (templates, properties, custom code etc). This data is stored in the 
 * MaiaConfig.xml file
 *
 * @author Julian Lyndon-Smith (julian+maia@dotr.com) 
 */

routine-level on error undo, throw.

class dotr.Maia.Crud.Project:

  /** temptable defs for the Project class itself */
  def temp-table TTProject no-undo xml-node-name "Project"     
    field ProjectName as char 
    field directory   as char /* table name of the class (base name only) [foo] */
    field StompServer as char 
    field Package     as char 
    
    index main is primary unique 
      ProjectName 
    .
  /** dataset so that we can read and write xml */          
  def dataset dsProject xml-node-name "Projects" for TTProject.

  
  def public property ConfigFile as char no-undo get . private set .
  
  def public property ProjectList as char no-undo 
    get():
      def var lv_List as char no-undo.
      
      for each TTProject no-lock:
        assign lv_list = lv_List + "," + TTProject.ProjectName.
      end.
      
      return trim(lv_list,",").
    end get. private set .

  constructor Project():
    
    if search("Maia.Properties") ne ? then LoadConfigFile("Maia.Properties").
    
    else 
    do:
      create TTProject.
      assign TTProject.ProjectName = "Sample"
             TTProject.Package     = "Sample"
             TTProject.Directory   = "Sample"
             TTProject.StompServer = '"localhost"'.
             
      assign ConfigFile = "./Maia.Properties".
      SaveDataSet().             
    end.
    
  end constructor .
  
  constructor Project(p_ConfigFile as char):
    LoadConfigFile(p_ConfigFile).
  end constructor .
  
  method private void SaveDataSet():
    def var lv_Data as longchar no-undo.
    
    dataset dsProject:WRITE-XML("longchar",lv_Data, yes  ).
      message string(lv_data) view-as alert-box.
    copy-lob from lv_data to file ConfigFile.
  end method.
  
  method private void LoadConfigFile(p_ConfigFile as char): 
    def var lv_Data as longchar no-undo.
    
    if search(p_ConfigFile) eq ? then undo, throw new AppError(substitute("Cannot read projects file &1",p_ConfigFile),0).
    
    /** check to see if it the object name, or the actual filename itself that has been passed in */
    assign file-info:file-name = search(p_ConfigFile).
    
    copy-lob from file file-info:full-pathname to lv_Data.
                              
    /** read the xml file into the temp-tables */
    dataset dsProject:READ-XML("longchar",lv_Data,"EMPTY",?,?).
    
    assign ConfigFile = file-info:full-pathname.
    
  end method.

  method public dotr.Maia.Model.Project New(p_ProjectName as char):
    def var Project1 as dotr.Maia.Model.Project no-undo.

    Project1 = new dotr.Maia.Model.Project(this-object).
    
    assign Project1:ProjectName = p_ProjectName.
    return Project1.                     
  end.
   
  method public dotr.Maia.Model.Project Get(p_ProjectName as char):
    def var Project1 as dotr.Maia.Model.Project no-undo.

    Project1 = new dotr.Maia.Model.Project(this-object).

    find first TTProject where TTProject.ProjectName eq p_ProjectName no-lock no-error.
    
    if not avail TTProject then return Project1.

    /** set properties of Project */
    assign Project1:Directory   = TTProject.directory
           Project1:Package     = TTProject.Package
           Project1:ProjectName = TTProject.ProjectName
           Project1:StompServer = TTProject.StompServer.
                     
    return Project1.                     
  end.

  /** save Project to xml file
   * @param p_Project : Project to save
   */

  method public void Save(p_Project as dotr.Maia.Model.Project):
    def var lv_mode as char.
    
    assign lv_mode = if can-find(TTproject where TTProject.ProjectName eq p_Project:ProjectName) 
                           then "CREATE"
                           else "update".
    
    WriteRecord(p_Project, lv_mode).
  end method.

  /** remove Project 
   * @param p_Project : Project to remove
   */
  method public void Remove(p_Project as dotr.Maia.Model.Project):
    WriteRecord(p_Project,"Remove").
  end method. 
  
  /** method to create or remove Project to xml file
   * @param p_Project : Project to save or remove
   * @param p_mode : remove or create
   */

  method private void WriteRecord(p_Project as dotr.Maia.Model.Project,p_Mode as char):
    
    repeat while true stop-after 5 transaction on error undo, throw
                                               on stop undo, leave:
      case p_Mode:
        when "Create" or when "update" then
        do:
          if p_mode eq "create" then create TTProject.
          
          else find first TTProject where TTProject.ProjectName eq p_Project:ProjectName exclusive no-error.
          
          assign TTProject.directory    = p_Project:Directory   
                 TTProject.Package      = p_Project:Package     
                 TTProject.ProjectName  = p_Project:ProjectName
                 TTProject.StompServe   = p_Project:StompServer.
        end.
        
        when "Remove" then
        do:
          find first TTProject where TTProject.ProjectName eq p_Project:ProjectName exclusive no-error.
          if avail TTProject then delete TTProject.
        end.
        
      end case.

      SaveDataSet().
      
      leave.
      
      catch e as Progress.Lang.Error :
        case e:GetMessageNum(1):
          when 132 then undo, throw new Progress.Lang.AppError(replace(replace(e:GetMessage(1),"**",""),"(132)",""),132). 
          
          otherwise     undo, throw new Progress.Lang.AppError(e:GetMessage(1),0).
        end case.
      end catch.
    end.
    
    return.
    
  end method.

  /** clears all temp-table record
   */
  
  method private void ClearTable():
    empty temp-table TTProject.
  end method.
      
end class.

